# Importing the Packages:
import optuna
import pandas as pd
import mlflow
import os

try:
    from sklearn.utils import safe_indexing
except ImportError:
    from sklearn.utils import _safe_indexing

    
from pycaret.classification import *
import numpy as np
import pandas as pd
from sklearn.metrics import recall_score





SENSORS_FEATURES = ['sensor_1','sensor_2','sensor_3','sensor_4','sensor_5',
 'sensor_6', 'sensor_7', 'sensor_8', 'sensor_9', 'sensor_10', 'sensor_11', 'sensor_12', 'sensor_13',
 'sensor_14', 'sensor_15', 'sensor_16', 'sensor_17', 'sensor_18', 'sensor_19', 'sensor_20', 'sensor_21', 'sensor_22', 'sensor_23', 'sensor_24']

STAGES = ['baseline', 'absorb', 'pause', 'desorb', 'flush']
TARGET_COL = 'result'

# artifact_path = 's3://raptor-mlflow-data/mlartifacts/1/83318437749c4e59a4950f55a98b2ad6/artifacts'
# artifact_path = 's3://raptor-mlflow-data'


def get_plots(model):

    model_attributes = dir(model)

    plot_model(model, plot = 'confusion_matrix', save=True)
    cm = os.path.join(os.getcwd(), 'Confusion Matrix.png')
    mlflow.log_artifact(cm)
    # log_artifact(cm)

    plot_model(model, plot = 'class_report', save=True) 
    cr = os.path.join(os.getcwd(), 'Class Report.png')
    mlflow.log_artifact(cr)  
    # log_artifact(cr)
    


    if 'predict_proba' in model_attributes:
        try:
            plot_model(model, plot = 'auc', save=True)
            auc = os.path.join(os.getcwd(), 'AUC.png')
            mlflow.log_artifact(auc) 

            plot_model(model, plot = 'pr', save=True) 
            pr = os.path.join(os.getcwd(), 'Precision Recall.png')
            mlflow.log_artifact(pr)

            plot_model(model, plot = 'calibration', save=True) 
            cc = os.path.join(os.getcwd(), 'Calibration Curve.png')
            mlflow.log_artifact(cc)
        except:
            pass
  

    
    if 'feature_importances_' in model_attributes:  
        plot_model(model, plot = 'feature', save=True)  
        fi = os.path.join(os.getcwd(), 'Feature Importance.png')
        mlflow.log_artifact(fi)





def specificity(actual, pred, pos_label=0):
    return recall_score(actual, pred, pos_label=pos_label)





def objective(trial, df, study_name, model_mode):

        mlflow.set_experiment(study_name)

       
        mlflow.set_tracking_uri(
            # "http://ec2-3-10-175-206.eu-west-2.compute.amazonaws.com:5000/"
            'http://ec2-3-10-210-150.eu-west-2.compute.amazonaws.com:5000/'
            )
        

        mlflow.start_run()


        features = df.drop('result', axis=1)
        clf = setup(df, target = 'result', silent=True, log_plots=True, experiment_name=study_name, preprocess=False)

        add_metric('specificity', 'Specificity', specificity)
        remove_metric('MCC')
        remove_metric('Kappa')

        MODELS = dict(models()['Name'])


        if model_mode == 'random':

            model_id = trial.suggest_categorical("model", list(MODELS.keys()))
            model_name = MODELS[model_id]
            model = create_model(model_id)
            predict_model(model)
            metrics = pull().iloc[:,1:]
            keys = metrics.columns
            values = metrics.values[0]
            metrics = dict(zip(keys, values))


            get_plots(model)
            mlflow.sklearn.log_model(model, artifact_path=model_name)
           
        # elif model_mode == 'xgb':

        #     model = create_model('xgb')
        #     predict_model(model)
        #     metrics = pull().iloc[:,1:]
        #     keys = metrics.columns
        #     values = metrics.values[0]
        #     metrics = dict(zip(keys, values))
            

        #     get_plots(model)
        #     mlflow.sklearn.log_model(model, artifact_path=model_name)
                    
        else:

            model = compare_models()

            model_name = model.__class__.__name__
            predict_model(model)
            metrics = pull().iloc[:,1:]
            keys = metrics.columns
            values = metrics.values[0]
            metrics = dict(zip(keys, values))


            get_plots(model)
            mlflow.sklearn.log_model(model, artifact_path=model_name)

            


            
        model_params = model.get_params()
        


        mlflow.log_metrics(metrics)

        try:
            mlflow.set_tag('sensor_cols', features)
  
        except:
            pass

        mlflow.set_tag('model_name', model_name)
        mlflow.set_tag('model_params', model_params)
        
        mlflow.set_tag('model_mode', model_mode)

        mlflow.end_run()







def train_experiments(df, study_name='raptor', direction='maximize', model_mode='random', n_trials=5):
    """trains several models during different trials and logs them. Experiemnt can be tracked on "http://ec2-3-10-175-206.eu-west-2.compute.amazonaws.com:5000/"


    Args:
        df (pandas dataframe): dataframe of cyclic sensor data to be used for training
        study_name (str, optional): optuna study name to use. Defaults to 'raptor'.
        direction (str, optional): dirtection of objective. Defaults to 'maximize'.
        stage (str or list, optional): Can take any of ('baseline', 'absorb', 'pause', 'desorb', 'flush') as a str or combination as a list. 
                                        Can also take str 'all' for all stages and 'random' for random stage or stages. Defaults to 'pause'.        sensor (str, optional): _description_. Defaults to '12'.
        use_average (bool, optional): if True, uses average of stages selected. Defaults to False.
        model_mode (str, optional): _description_. Defaults to 'random'.
        n_trials (int, optional): number of times to run experiments. Defaults to 5.
        suggest (bool, optional): if True, suggest features, stage, model mode to use for training. Defaults to False.
    """    
    
    

    study = optuna.create_study(study_name=study_name, direction=direction)
    study.optimize(lambda trial: objective(trial, df, study_name, model_mode), n_trials=n_trials)

    print('Click on this link to track experiments: ', 
    # "http://ec2-3-10-175-206.eu-west-2.compute.amazonaws.com:5000/"
    'http://ec2-3-10-210-150.eu-west-2.compute.amazonaws.com:5000/'
    )





from scipy.signal import find_peaks
from scipy import stats
import pandas as pd
import numpy as np
import xgboost as xgb
from boruta import BorutaPy
from tsfresh.utilities.dataframe_functions import impute
from tsfresh.feature_extraction import ComprehensiveFCParameters
from tsfresh import extract_features, extract_relevant_features, select_features





FEATURES = [ 'exp_unique_id', 'timesteps', 'sensor_1', 'sensor_2',
       'sensor_3', 'sensor_4', 'sensor_5', 'sensor_6', 'sensor_7', 'sensor_8',
       'sensor_9', 'sensor_10', 'sensor_11', 'sensor_12', 'sensor_13',
       'sensor_14', 'sensor_15', 'sensor_16', 'sensor_17', 'sensor_18',
       'sensor_19', 'sensor_20', 'sensor_21', 'sensor_22', 'sensor_23',
       'sensor_24', 'result']

extraction_settings = ComprehensiveFCParameters()


def get_features(df, unique_id='exp_unique_id', label='result', timesteps='timesteps', features=FEATURES):


    df = df[FEATURES]

    X = df.drop(label, axis=1)
    y = df.groupby(unique_id).first()[label]

    X = extract_features(X, column_id=unique_id, column_sort=timesteps,
                     default_fc_parameters=extraction_settings,
                     
                     # we impute = remove all NaN features automatically
                     impute_function=impute)



    return X, y

    

forest = xgb.XGBClassifier()
boruta = BorutaPy(
estimator = forest, 
n_estimators = 'auto',
max_iter = 100 # number of trials to perform
)

def get_relevant_features(X, y):

    ### fit Boruta (it accepts np.array, not pd.DataFrame)
    boruta.fit(np.array(X), np.array(y))
    
    
    ### print results
    green_area = X.columns[boruta.support_].to_list()
    blue_area = X.columns[boruta.support_weak_].to_list()

    relevant_features = green_area + blue_area
    X_f = X[relevant_features]
    dff = X_f.join(y)

    dff.columns = dff.columns.str.replace('[",\d, (, )]', '')

    return dff, relevant_features



import pandas as pd
import os
import random
import numpy as np
import re
import warnings




SENSOR_FEATURES = ['sensor_1','sensor_2','sensor_3','sensor_4','sensor_5',
 'sensor_6', 'sensor_7', 'sensor_8', 'sensor_9', 'sensor_10', 'sensor_11', 'sensor_12', 'sensor_13',
 'sensor_14', 'sensor_15', 'sensor_16', 'sensor_17', 'sensor_18', 'sensor_19', 'sensor_20', 'sensor_21', 'sensor_22', 'sensor_23', 'sensor_24']



def find_header_value(field, filename):
    """Finds the location of a keyword (exp_name, date, etc.) from the header of the .txt file

    Args:
        field (_type_): _description_
        filename (_type_): _description_

    Returns:
        _type_: _description_
    """
    
  
  
    with open(filename) as f:
        for line_no, line in enumerate(f): 
            if line.startswith(field):
                # Get the part of the string after the field name
                end_of_string = line[len(field):]
                string = end_of_string[:-1]
                # break
                return string, line_no



def get_exp_stage(time, data_points_per_sec=4, baseline=2, absorb=7, pause=1, desorb=5, flush=23):
    
    # data_points_per_sec: data point per second

    baseline_time = baseline*data_points_per_sec
    if time <= baseline_time:
        return 'baseline'
    absorb_time = baseline_time + absorb*data_points_per_sec
    if time <= absorb_time:
        return 'absorb'
    pause_time = absorb_time + pause*data_points_per_sec
    if time <= pause_time:
        return 'pause'
    desorb_time = pause_time + desorb*data_points_per_sec
    if time <= desorb_time:
        return 'desorb'
    flush_time = desorb_time + flush*data_points_per_sec
    if time <= flush_time:
        return 'flush'
    wait_time = flush_time + flush*data_points_per_sec
    if time <= wait_time:
        return 'wait'
    else:
        warnings.warn("Warning...........Can't process stage")
        return 'nan'
    


def rename_columns(df):


    for col in df.columns:
        if 'Sen' in col:
            new_col = re.findall('\d+', col)[0]
            df.rename(columns={col:f'sensor_{new_col}'}, inplace=True)
    df.rename(columns={'Data Points': 'timesteps'}, inplace=True)
    df.columns = df.columns.str.replace(r"[( r.h. )]", "")

    return df




def get_label(f):   

    filename = f.split('/')[-1]
    exp_name, _ = find_header_value('Name of the experiment =', f)

    if ("Neg" in filename) or ("Control" in filename) or ('CONTROL' in filename) or ("Neg" in exp_name) or ("Control" in exp_name) or ('CONTROL' in exp_name) or ("Neg" in filename):
        return "Control"
    elif ("Pos" in filename) or ("Covid" in filename) or ('COVID' in filename) or ("Pos" in exp_name) or ("Covid" in exp_name) or ('COVID' in exp_name) or ("Pos" in filename):
        return "Covid"
    else:
        warnings.warn("Warning...........Can't get label")
        return 'Other'
    

def get_data_points_per_sec(f):
    global data_points_per_sec
    field = 'Acquired data point per second = '
    data_points_per_sec = float(find_header_value(field, f)[0])
    return data_points_per_sec


def get_exp_stage_duration(f):

    global baseline, absorb, pause, desorb, flush

    baseline = float(find_header_value('BaseLine = ', f)[0])
    absorb = float(find_header_value('Absorb = ', f)[0])
    pause = float(find_header_value('Pause = ', f)[0])
    desorb = float(find_header_value('Desorb = ', f)[0])
    flush = float(find_header_value('Flush = ', f)[0])

    return baseline, absorb, pause, desorb, flush



def offset_one_sample(df_temp):

    datapoints = baseline * data_points_per_sec
    avg_baseline = df_temp[SENSOR_FEATURES].iloc[:datapoints].mean().values

    df_temp[SENSOR_FEATURES] = df_temp[SENSOR_FEATURES].apply(lambda row: row - avg_baseline, axis=1)

    return df_temp


def offset_batch_samples(df, id_col='exp_unique_id'):
    """normalises all sensors in each cycle of experiment by subtracting the average of the sensor voltage in the 1st 2 sec from sensor voltage in other periods 

    Args:
        df (pandas dataframe): dataframe of cyclic sensor data
        sec (int, optional): number of seconds to use for offsetting. Defaults to 2.
        id_col (str, optional): unique id of each cycle of experiment. Defaults to 'exp_unique_id'.

    Returns:
        pandas dataframe: _description_
    """
    samples = df.groupby(id_col)
    df_list = []
    for i, sample in samples:
        df_list.append(offset_one_sample(sample))
    return pd.concat(df_list)



def fillna_columns(df_temp):

    num = df_temp._get_numeric_data()
    num[num < 0] = np.nan
    num = num.fillna(num.mean())
    return df_temp


def preprocess_single_file(f, parse_time=True, parse_filename=True, rename_column=True, fillna=True, offset=True, has_label=True):
    """preprocess a single cycle (one txt file) of experiment

    Args:
        f (str): file path with txt extension
        rename_column (bool, optional): rename columns. Defaults to True.
        fillna (bool, optional): replaces extreme negative values with mean of column. Defaults to True.
        offset (bool, optional): normalises all sensors in each cycle of experiment by subtracting the average of the sensor voltage in the 1st 2 sec from sensor voltage in other periods . Defaults to True.
        has_label (bool, optional): extracts label and add to dataframe. Defaults to True.

    Returns:
        pandas dataframe: dataframe of features in a cycle(one txt file)
    """   

    _, line_no = find_header_value('Data Points', f)


    df_temp = pd.read_csv(f,sep='\t',header=(line_no - 1))

    data_points_per_sec = get_data_points_per_sec(f)
    baseline, absorb, pause, desorb, flush = get_exp_stage_duration(f)


    df_temp['measurement_stage'] = df_temp['Data Points'].apply(get_exp_stage, data_points_per_sec=data_points_per_sec, 
                                                                baseline=baseline, absorb=absorb, pause=pause, 
                                                                desorb=desorb, flush=flush)

    
    if parse_time:
      date, _ = find_header_value('Date = ', f)
      df_temp['date_exp'], _ = find_header_value('Date = ', f)
      df_temp['time_elapsed'] = df_temp.index / 4
      time_start, _ = find_header_value('Time = ', f)
      time_elapsed = df_temp.index / 4
      timestamp = pd.to_datetime(date + " " + time_start)
      df_temp['datetime_exp_start'] = time_start
      df_temp['datetime_exp'] = pd.to_datetime(date + " " + time_start) + pd.to_timedelta(time_elapsed, unit='s')


    repeat_no = find_header_value('Repeat No. = ', filename)[0].split('/')[0]
    df_temp['repeat_no'] = repeat_no
    
    if parse_filename:
      df_temp['filename'] = f.split('/')[-1]
    
    if has_label:
        df_temp['result'] = get_label(f)

    
    df_temp['exp_name'] = find_header_value('Name of the experiment = ', f)[0][1:-1]

    

    if rename_column:
      df = rename_columns(df_temp)


    if fillna:
        df_temp = fillna_columns(df_temp)


    if offset:
        df_temp = offset_one_sample(df_temp)


    return df_temp

def get_all_files(parent_dir):
    all_files = []
    for path, subdirs, files in os.walk(parent_dir):
        for name in files:
            if (name.endswith('txt')) and (len(name) > 25) and ('Wash' not in name):
                all_files.append(os.path.join(path, name))
    return all_files

def get_single_repeat(df):
    """select one (chosen at random) of repeat experiemnts

    Args:
        df (pandas dataframe): dataframe of cyclic sensor data with repeat experiments

    Returns:
        pandas dataframe: dataframe with only one of each repeat experiemnt
    """

    g = df.groupby(['exp_name'])

    exp_names = list(set(df['exp_name']))
    all = []
    for i in exp_names:
        df = g.get_group(i)

        uniq_files = list(df['filename'].unique())
        file_chosen = random.choice(uniq_files)
        df = df[df['filename'] == file_chosen]
        all.append(df)
    return pd.concat(all)


def reorder_columns(df):
    """reorders features with exp_unique_id as first column asnd result as last

    Args:
        df (pandas dataframe): _description_

    Returns:
        pandas dataframe: _description_
    """

    col_list = df.columns.tolist()
    new_col_list = [col_list[-1]] + [col_list[-2]]  + col_list[:-2] 
    df = df[new_col_list]

    return df

def preprocess_all_files(path_to_measurements, parse_time=True, parse_filename=True, rename_column=True, no_repeat=False, fillna=True, offset=False, has_label=True):
    """preprocess all cycle (all txt file) of experiment

    Args:
        path_to_measurements (str): folder path to measurements
        parse_time (bool, optional): extract time features. Defaults to True.
        parse_filename (bool, optional): adds filename to dataframe. Defaults to True.
        rename_column (bool, optional): rename columns. Defaults to True.
        no_repeat (bool, optional): _description_. Defaults to False.
        fillna (bool, optional): replaces extreme negative values with mean of column. Defaults to True.
        offset (bool, optional): normalises all sensors in each cycle of experiment by subtracting the average of the sensor voltage in the 1st 2 sec from sensor voltage in other periods . Defaults to True.
        has_label (bool, optional): extracts label and add to dataframe. Defaults to True.

    Returns:
        pandas dataframe: dataframe of features in a cycle(one txt file)
    """


    list_files = get_all_files(path_to_measurements)


    df_list = []
    for i, f in enumerate(list_files):

        df_temp = preprocess_single_file(f, parse_time=parse_time, parse_filename=parse_filename, offset=offset, fillna=fillna, has_label=has_label)
        df_temp['exp_unique_id'] = i
        df_list.append(df_temp)

    
    df = pd.concat(df_list)
    df = reorder_columns(df)


    if no_repeat:
        return get_unique_files(df)
    else:
        return df








  




import joblib
from io import BytesIO
import boto3















def load_model(path):
    ''' 
       Function to load a joblib file from an s3 bucket or local directory.
       Arguments:
       * path: an s3 bucket or local directory path where the file is stored
       Outputs:
       * file: Joblib file loaded
    '''

    # Path is an s3 bucket
    if path[:5] == 's3://':
        s3_bucket, s3_key = path.split('/')[2], path.split('/')[3:]
        s3_key = '/'.join(s3_key)
        with BytesIO() as f:
            boto3.client("s3").download_fileobj(Bucket=s3_bucket, Key=s3_key, Fileobj=f)
            f.seek(0)
            file = joblib.load(f)
    
    # Path is a local directory 
    else:
        with open(path, 'rb') as f:
            file = joblib.load(f)
    
    return file









def load_ml_model(path_to_model):
  model = pickle.load(open(path_to_model, "rb"))
  return model

def random_sample(df_list,x_columns):
  i = random.randint(0, len(df_list))
  df_temp = df_list[i]
  df_temp = df_temp[x_columns]
  # df_temp = df_temp.iloc[:,[1,2,3,4,5,6,7,8,9,10,11,12]]
  j = random.randint(0, len(df_temp.count()))
  input_data_sample = df_temp.iloc[[j]]
  print(input_data_sample)
  # 
  x = input_data_sample.columns.tolist()
  y = input_data_sample.iloc[0].values.tolist()
  # 
  plt.bar(x, y)
  fig = plt.gcf()
  fig.autofmt_xdate()
  # 
  return input_data_sample

def sample_prediction(input_data_sample, model):
  # 
  print('Prediction: ' + str(model.predict(input_data_sample)))
  print('Prediction Probability: ' + str(model.predict_proba(input_data_sample)))


import pandas as pd
import warnings


def get_data(data='validated_breath_data'):
    """Loads data into a dataframe

    Args:
        data (str, optional): name of data to be loaded. 
        examples of data include 'validated_breath_data', 'lab_data', 'repeat_experiment', 'handheld_data'.
        Defaults to 'validated_breath_data'.


    Returns:
        pandas dataframe: 
    """



    if data == 'validated_breath_data':
        url = 'https://drive.google.com/file/d/1CJ3e8NKS2KgUidZMLBFP0ktq2_y87eHb/view?usp=sharing'
        url='https://drive.google.com/uc?id=' + url.split('/')[-2]


    elif data == 'lab_data':
        url = 'https://drive.google.com/file/d/1dz3vrsOLbqiedSIslxbioUhS82MB96WS/view?usp=sharing' 
        url='https://drive.google.com/uc?id=' + url.split('/')[-2]


    elif data == 'repeat_experiment':
        url = 'https://drive.google.com/file/d/1r83aSt_LJ-nGVLXohFKbk8oUlGwtJ4Gi/view?usp=sharing'
        url='https://drive.google.com/uc?id=' + url.split('/')[-2]



    elif data == 'handheld_data':
        url = 'https://drive.google.com/file/d/1YZ6PDQpT9QkeVBLuRK6hThgaI54av6sf/view?usp=sharing'
        url='https://drive.google.com/uc?id=' + url.split('/')[-2]


    else:

        return warnings.warn("Warning...........Data not available. Available data include: \
        'validated_breath_data', 'lab_data', 'repeat_experiment', 'handheld_data'")

        

    return pd.read_csv(url, index_col=0)